@@ -3,13 +3,14 @@
 A sas7bdat reader. File format taken from
 https://github.com/BioStatMatt/sas7bdat/blob/master/inst/doc/sas7bdat.rst
 """
-import os
-import sys
 import csv
-import struct
 import locale
 import logging
+import math
+import os
 import platform
+import struct
+import sys
 from cStringIO import StringIO
 from datetime import datetime, timedelta
 from collections import namedtuple
@@ -56,35 +57,35 @@ class SAS7BDAT(object):
             "\xb3\x14\x11\xcf\xbd\x92\x08\x00\x09\xc7\x31\x8c\x18\x1f\x10\x11"
 
     # Host systems known to work
-    KNOWNHOSTS = set(["WIN_PRO", "WIN_NT", "WIN_NTSV", "WIN_SRV", "WIN_ASRV",
-                      "XP_PRO", "XP_HOME", "NET_ASRV", "NET_DSRV", "NET_SRV",
-                      "WIN_98", "W32_VSPRO", "WIN", "WIN_95", "X64_VSPRO",
-                      "AIX", "X64_ESRV", "W32_ESRV", "W32_7PRO", "W32_VSHOME",
-                      "X64_7HOME", "X64_7PRO", "X64_SRV0", "W32_SRV0",
-                      "X64_ES08", "Linux", "HP-UX"])
+    KNOWNHOSTS = {"WIN_PRO", "WIN_NT", "WIN_NTSV", "WIN_SRV", "WIN_ASRV",
+                  "XP_PRO", "XP_HOME", "NET_ASRV", "NET_DSRV", "NET_SRV",
+                  "WIN_98", "W32_VSPRO", "WIN", "WIN_95", "X64_VSPRO",
+                  "AIX", "X64_ESRV", "W32_ESRV", "W32_7PRO", "W32_VSHOME",
+                  "X64_7HOME", "X64_7PRO", "X64_SRV0", "W32_SRV0",
+                  "X64_ES08", "Linux", "HP-UX"}
 
     # Subheader signatures, 32 and 64 bit, little and big endian
-    SUBH_ROWSIZE = set(["\xF7\xF7\xF7\xF7", "\x00\x00\x00\x00\xF7\xF7\xF7\xF7",
-                        "\xF7\xF7\xF7\xF7\x00\x00\x00\x00"])
-    SUBH_COLSIZE = set(["\xF6\xF6\xF6\xF6", "\x00\x00\x00\x00\xF6\xF6\xF6\xF6",
-                        "\xF6\xF6\xF6\xF6\x00\x00\x00\x00"])
-    SUBH_COLTEXT = set(["\xFD\xFF\xFF\xFF", "\xFF\xFF\xFF\xFD",
-                        "\xFD\xFF\xFF\xFF\xFF\xFF\xFF\xFF",
-                        "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFD"])
-    SUBH_COLATTR = set(["\xFC\xFF\xFF\xFF", "\xFF\xFF\xFF\xFC",
-                        "\xFC\xFF\xFF\xFF\xFF\xFF\xFF\xFF",
-                        "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFC"])
-    SUBH_COLNAME = set(["\xFF\xFF\xFF\xFF",
-                        "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"])
-    SUBH_COLLABS = set(["\xFE\xFB\xFF\xFF", "\xFF\xFF\xFB\xFE",
-                        "\xFE\xFB\xFF\xFF\xFF\xFF\xFF\xFF",
-                        "\xFF\xFF\xFF\xFF\xFF\xFF\xFB\xFE"])
-    SUBH_COLLIST = set(["\xFE\xFF\xFF\xFF", "\xFF\xFF\xFF\xFE",
-                        "\xFE\xFF\xFF\xFF\xFF\xFF\xFF\xFF",
-                        "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFE"])
-    SUBH_SUBHCNT = set(["\x00\xFC\xFF\xFF", "\xFF\xFF\xFC\x00",
-                        "\x00\xFC\xFF\xFF\xFF\xFF\xFF\xFF",
-                        "\xFF\xFF\xFF\xFF\xFF\xFF\xFC\x00"])
+    SUBH_ROWSIZE = {"\xF7\xF7\xF7\xF7", "\x00\x00\x00\x00\xF7\xF7\xF7\xF7",
+                    "\xF7\xF7\xF7\xF7\x00\x00\x00\x00"}
+    SUBH_COLSIZE = {"\xF6\xF6\xF6\xF6", "\x00\x00\x00\x00\xF6\xF6\xF6\xF6",
+                    "\xF6\xF6\xF6\xF6\x00\x00\x00\x00"}
+    SUBH_COLTEXT = {"\xFD\xFF\xFF\xFF", "\xFF\xFF\xFF\xFD",
+                    "\xFD\xFF\xFF\xFF\xFF\xFF\xFF\xFF",
+                    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFD"}
+    SUBH_COLATTR = {"\xFC\xFF\xFF\xFF", "\xFF\xFF\xFF\xFC",
+                    "\xFC\xFF\xFF\xFF\xFF\xFF\xFF\xFF",
+                    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFC"}
+    SUBH_COLNAME = {"\xFF\xFF\xFF\xFF",
+                    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"}
+    SUBH_COLLABS = {"\xFE\xFB\xFF\xFF", "\xFF\xFF\xFB\xFE",
+                    "\xFE\xFB\xFF\xFF\xFF\xFF\xFF\xFF",
+                    "\xFF\xFF\xFF\xFF\xFF\xFF\xFB\xFE"}
+    SUBH_COLLIST = {"\xFE\xFF\xFF\xFF", "\xFF\xFF\xFF\xFE",
+                    "\xFE\xFF\xFF\xFF\xFF\xFF\xFF\xFF",
+                    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFE"}
+    SUBH_SUBHCNT = {"\x00\xFC\xFF\xFF", "\xFF\xFF\xFC\x00",
+                    "\x00\xFC\xFF\xFF\xFF\xFF\xFF\xFF",
+                    "\xFF\xFF\xFF\xFF\xFF\xFF\xFC\x00"}
 
     # Page types
     PAGE_META = 0
@@ -337,13 +338,13 @@ class SAS7BDAT(object):
             # Timestamp is epoch 01/01/1960
             datecreated = self.readVal('d', h, 164 + align1, 8)
             try:
-                datecreated = datetime.strptime('1960/01/01', "%Y/%m/%d") +\
+                datecreated = datetime(1960, 1, 1) +\
                     timedelta(seconds=datecreated)
             except:
                 pass
             datemodified = self.readVal('d', h, 172 + align1, 8)
             try:
-                datemodified = datetime.strptime('1960/01/01', "%Y/%m/%d") + \
+                datemodified = datetime(1960, 1, 1) + \
                     timedelta(seconds=datemodified)
             except:
                 pass
@@ -390,7 +391,7 @@ class SAS7BDAT(object):
             colattr = []
             colname = []
             collabs = []
-            data = []
+            # data = []
             for x in self.readSubheaders(f, pagecount, pagesize):
                 if x is None:
                     continue
@@ -529,38 +530,37 @@ class SAS7BDAT(object):
         except:
             locale.setlocale(locale.LC_NUMERIC, '')
         places = max(0, places)
-        return locale.format('%%f' % places, num, True)
+        return locale.format('%%.%df' % places, num, True)
 
     def formatValue(self, val, fmt):
         if fmt in self.formatters:
             return self.formatters[fmt](val)
-        noFormat = set(['', '$', '$CHAR', '$F', 'BEST', 'F', 'NLNUM',
-                        'SPECFMT'])
+        noFormat = {'', '$', '$CHAR', '$F', 'BEST', 'F', 'NLNUM', 'SPECFMT'}
         if fmt is None:
             fmt = ''
         fmt = fmt.upper()
         if not fmt or fmt in noFormat:
             return val
         elif fmt == 'MMDDYY':
-            return (datetime.strptime('1960/01/01', "%Y/%m/%d") +
+            return (datetime(1960, 1, 1) +
                     timedelta(days=val)).strftime('%m/%d/%Y')
         elif fmt == 'DDMMYY':
-            return (datetime.strptime('1960/01/01', "%Y/%m/%d") +
+            return (datetime(1960, 1, 1) +
                     timedelta(days=val)).strftime('%d.%m.%Y')
         elif fmt == 'JULIAN':
-            return (datetime.strptime('1960/01/01', "%Y/%m/%d") +
+            return (datetime(1960, 1, 1) +
                     timedelta(days=val)).strftime('%Y%d')
         elif fmt == 'MONYY':
-            return (datetime.strptime('1960/01/01', "%Y/%m/%d") +
+            return (datetime(1960, 1, 1) +
                     timedelta(days=val)).strftime('%b %Y').upper()
         elif fmt == 'DATE':
-            return (datetime.strptime('1960/01/01', "%Y/%m/%d") +
+            return (datetime(1960, 1, 1) +
                     timedelta(days=val)).strftime('%d%b%Y').upper()
         elif fmt == 'DATETIME':
-            return (datetime.strptime('1960/01/01', "%Y/%m/%d") +
+            return (datetime(1960, 1, 1) +
                     timedelta(seconds=val)).strftime('%Y-%m-%d %H:%M:%S')
         elif fmt == 'TIME':
-            return (datetime.strptime('1960/01/01', "%Y/%m/%d") +
+            return (datetime(1960, 1, 1) +
                     timedelta(seconds=val)).strftime('%H:%M:%S')
         elif fmt == 'COMMA':
             if isinstance(val, int):
@@ -683,15 +683,23 @@ class SAS7BDAT(object):
                                 if col.attr.type == 'character':
                                     val = self.readVal('s', raw, 0,
                                                        col.attr.length)
-                                    val = val.lstrip().strip()
+                                    val = val.strip()
                                 else:
                                     val = self.readVal(col.attr.type, raw, 0,
                                                        col.attr.length)
-                                val = self.formatValue(val, col.label.format)
+
+                                if isinstance(val, float) and math.isnan(val):
+                                    val = None
+                                else:
+                                    val = self.formatValue(
+                                        val, col.label.format)
                             except KeyboardInterrupt:
                                 return
-                            except:
-                                break
+                            except Exception as e:
+                                self.logger.error(
+                                    'Exception parsing col %s:\n%s',
+                                    str(col), str(e))
+                                raise
                             row.append(val)
                     base += self.header.rowlength
                     if row:
@@ -710,9 +718,11 @@ class SAS7BDAT(object):
                              delimiter=delimiter)
             i = 0
             for i, line in enumerate(self.readData(), 1):
-                if not line:
-                    i -= 1
-                    continue
+                if len(line) != self.header.colcount:
+                    self.logger.error(
+                        'ERROR. Parsed line into %d columns, expected %d.\n%s',
+                        len(line), self.header.colcount, line)
+                    break
                 if not i % stepSize:
                     self.logger.info('%.1f%% complete',
                                      float(i) / self.header.rowcount * 100.0)
@@ -728,6 +738,11 @@ class SAS7BDAT(object):
             if outF is not None:
                 outF.close()
 
+    def toDataFrame(self):
+        import pandas as pd
+        data = list(self.readData())
+        return pd.DataFrame([dict(zip(data[0], x)) for x in data[1:]])
+
 
 if __name__ == '__main__':
     pass  # TODO: write some unit tests
